// Deobjectivizer - The program that makes C bindings to ObjC frameworks
// Copyright (C) 2010 Charles E. Davis

// This program is free software...

// Deobjectivizer.cpp - Implements the Deobjectivizer class, which takes an
//                      Objective-C header and produces a C binding to it.

#include <clang/AST/AST.h>
#include <clang/AST/ASTConsumer.h>
#include <clang/AST/ParentMap.h>
#include <clang/Basic/Diagnostic.h>
#include <clang/Basic/IdentifierTable.h>
#include <clang/Basic/SourceManager.h>
#include <clang/Basic/TargetInfo.h>
#include <clang/Lex/Lexer.h>
#include <clang/Rewrite/Rewriter.h>
#include <llvm/ADT/SmallVector.h>
#include <llvm/ADT/StringRef.h>
#include <llvm/ADT/Triple.h>
#include <llvm/Support/raw_ostream.h>
#include <llvm/Support/MemoryBuffer.h>
#include <llvm/Support/PathV2.h>

namespace {
  // Based largely off clang's RewriteObjC playground.
  class Deobjectivizer: public clang::ASTConsumer {
    clang::Rewriter rewriter;
    clang::Diagnostic& diag;
    const clang::LangOptions& lang_opts;
    unsigned rewrite_failed_diag;

    clang::ASTContext* ctx;
    clang::SourceManager* sm;
    clang::TranslationUnitDecl* tu_decl;
    clang::FileID main_id;
    const char* main_start,* main_end;
    clang::SourceLocation last_inc;
    std::string fwk_name;
    std::string c_fwk_name;

    clang::FunctionDecl* SelRegisterNameFunctionDecl;

    std::string in_file_name;
    llvm::raw_ostream* out_file;    // The header file
    llvm::raw_ostream* out_c_file;  // The implementation file
    std::string out_c_cont;
    llvm::raw_string_ostream* out_c_buf; // Impl. file's buffer

    bool silence_warnings;

    std::string preamble;

    // This maps an original source AST to it's rewritten form. This allows
    // us to avoid rewriting the same node twice (which is very uncommon).
    // This is needed to support some of the exotic property rewriting.
    llvm::DenseMap<clang::Stmt *, clang::Stmt *> ReplacedNodes;

    clang::FunctionDecl *CurFunctionDef;
    clang::VarDecl *GlobalVarDecl;

    void RewriteFrameworkInclude(const char*& buf, clang::SourceLocation start);

  public:
    virtual void Initialize(clang::ASTContext& context);

    virtual void HandleTopLevelDecl(clang::DeclGroupRef D) {
      for(clang::DeclGroupRef::iterator I = D.begin(), E = D.end();
          I != E;
          ++I)
        HandleTopLevelSingleDecl(*I);
    }
    void HandleTopLevelSingleDecl(clang::Decl* D);

    Deobjectivizer(std::string fwkName, std::string inFile,
                   llvm::raw_ostream* OS, llvm::raw_ostream* C_OS,
                   clang::Diagnostic& D, const clang::LangOptions& LOpts,
                   bool silenceWarn);

    ~Deobjectivizer() {}

    virtual void HandleTranslationUnit(clang::ASTContext& C);

    void ReplaceStmt(clang::Stmt* old, clang::Stmt* new_, int size = 0) {
      // If replacement succeeded or warning disabled return with no
      // warning.
      if(!rewriter.ReplaceStmt(old, new_) || silence_warnings)
        return;
      diag.Report(ctx->getFullLoc(old->getLocStart()),
                  rewrite_failed_diag)
        << old->getSourceRange();
    }
    void ReplaceStmtWithRange(clang::Stmt* old, clang::Stmt* new_,
                              clang::SourceRange range) {
      // Measure the old text.
      int size = rewriter.getRangeSize(range);
      if(size == -1) {
        diag.Report(ctx->getFullLoc(old->getLocStart()),
                    rewrite_failed_diag)
          << old->getSourceRange();
        return;
      }
      // Get the new text.
      std::string str;
      llvm::raw_string_ostream stm(str);
      new_->printPretty(stm, *ctx, 0, clang::PrintingPolicy(lang_opts));
      // If replacement succeeded or warning disabled return with no
      // warning.
      if(!rewriter.ReplaceText(range.getBegin(), size, stm.str()) ||
         silence_warnings)
        return;
      diag.Report(ctx->getFullLoc(old->getLocStart()),
                  rewrite_failed_diag)
        << old->getSourceRange();
    }
    void InsertText(clang::SourceLocation loc, const char* data,
                    unsigned len, bool after = true) {
      // If insertion succeeded or warning disabled return with no
      // warning.
      if(!rewriter.InsertText(loc, llvm::StringRef(data, len), after) ||
         silence_warnings)
        return;
      diag.Report(ctx->getFullLoc(loc), rewrite_failed_diag);
    }
    void RemoveText(clang::SourceLocation loc, unsigned len) {
      // You know the drill.
      if(!rewriter.RemoveText(loc, len) || silence_warnings)
        return;
      diag.Report(ctx->getFullLoc(loc), rewrite_failed_diag);
    }
    void ReplaceText(clang::SourceLocation start, unsigned orig_len,
                     const char* str, unsigned len) {
      if(!rewriter.ReplaceText(start, orig_len,
                               llvm::StringRef(str, len)) ||
         silence_warnings)
        return;
      diag.Report(ctx->getFullLoc(start), rewrite_failed_diag);
    }

    llvm::StringRef getCFrameworkName(void) {
      return c_fwk_name;
    }

    // Syntactic rewriting.
    void RewriteInclude(void);
    void RewriteAtEnd(clang::ObjCContainerDecl* cont);
    void RewriteForwardClassDecl(clang::ObjCClassDecl* decl);
    void RewriteInterfaceDecl(clang::ObjCInterfaceDecl* decl);
    void RewriteCategoryDecl(clang::ObjCCategoryDecl* decl);
    void RewriteProtocolDecl(clang::ObjCProtocolDecl* decl);
    void RewriteForwardProtocolDecl(clang::ObjCForwardProtocolDecl* decl);
    void RewriteMethodDeclaration(clang::ObjCMethodDecl* meth);
    void RewriteProperty(clang::ObjCPropertyDecl* prop);
    void RewriteFunctionDecl(clang::FunctionDecl* func);
    void RewriteRecordBody(clang::RecordDecl* rec);
    void RewriteObjCQualifiedInterfaceTypes(clang::Decl* decl);
    void RewriteObjCQualifiedInterfaceTypes(clang::Expr* expr);

    // Expression rewriting.
    clang::Stmt* RewriteAtEncode(clang::ObjCEncodeExpr* expr);
    clang::Stmt* RewriteAtSelector(clang::ObjCSelectorExpr* expr);

    // Syntax and expression synthesis.
    clang::CallExpr* SynthesizeCallToFunctionDecl(clang::FunctionDecl* func,
                                                  clang::Expr** args,
                                                  unsigned nargs);
    void SynthSelRegisterNameFunctionDecl(void);

    // Miscellaneous utility.
    bool BufferContainsIfPPDirective(const char* start, const char* end);
    bool BufferContainsIfDefPPDirective(const char* start, const char* end);
    bool BufferContainsIfNdefPPDirective(const char* start, const char* end);
    bool BufferContainsEndIfPPDirective(const char* start, const char* end);

    bool isObjCType(clang::QualType T) {
      if(!lang_opts.ObjC1 && !lang_opts.ObjC2) return false;

      clang::QualType OCT = ctx->getCanonicalType(T).getUnqualifiedType();

      if(OCT == ctx->getCanonicalType(ctx->getObjCIdType()) ||
         OCT == ctx->getCanonicalType(ctx->getObjCClassType()))
        return true;
      if(const clang::PointerType* PT = OCT->getAs<clang::PointerType>()) {
        if(isa<clang::ObjCInterfaceType>(PT->getPointeeType()) ||
           PT->getPointeeType()->isObjCQualifiedIdType())
          return true;
      }
      return false;
    }
    void RewriteCastExpr(clang::CStyleCastExpr* expr);
  };
  clang::CStyleCastExpr* NoTypeInfoCStyleCastExpr(clang::ASTContext* ctx,
                                                  clang::QualType type,
                                                  clang::CastKind kind,
                                                  clang::Expr* e) {
    clang::TypeSourceInfo* tinfo = ctx->getTrivialTypeSourceInfo(type,
                                                                 clang::SourceLocation());
    return clang::CStyleCastExpr::Create(*ctx, type, clang::VK_RValue,
                                         kind, e, 0, tinfo,
                                         clang::SourceLocation(),
                                         clang::SourceLocation());
  }
}

Deobjectivizer::Deobjectivizer(std::string fwkName, std::string inFile,
                               llvm::raw_ostream* OS, llvm::raw_ostream* C_OS,
                               clang::Diagnostic& D,
                               const clang::LangOptions& LOpts,
                               bool silenceWarn)
: diag(D), lang_opts(LOpts), fwk_name(fwkName), in_file_name(inFile),
out_file(OS), out_c_file(C_OS), out_c_cont(),
out_c_buf(new llvm::raw_string_ostream(out_c_cont)),
silence_warnings(silenceWarn) {
  c_fwk_name = fwk_name + "C";
  rewrite_failed_diag = diag.getCustomDiagID(clang::Diagnostic::Warning,
    "rewriting sub-expression within a macro (may not be correct)");
 }

clang::ASTConsumer* CreateDeobjectivizer(const std::string& fwk_name,
                                         const std::string& in_file,
                                         llvm::raw_ostream* os,
                                         llvm::raw_ostream* c_os,
                                         clang::Diagnostic& diag,
                                         const clang::LangOptions& lopts,
                                         bool silence_warn) {
  return new Deobjectivizer(fwk_name, in_file, os, c_os, diag, lopts,
                            silence_warn);
}

void Deobjectivizer::Initialize(clang::ASTContext& context) {
  ctx = &context;
  sm = &ctx->getSourceManager();
  tu_decl = ctx->getTranslationUnitDecl();
  CurFunctionDef = NULL;
  GlobalVarDecl = NULL;

  // Get the ID and start/end of the main file.
  main_id = sm->getMainFileID();
  const llvm::MemoryBuffer* main_buf = sm->getBuffer(main_id);
  main_start = main_buf->getBufferStart();
  main_end = main_buf->getBufferEnd();

  rewriter.setSourceMgr(ctx->getSourceManager(), ctx->getLangOptions());

  preamble = "#pragma once\n";
  preamble += "#include <Deobjectivizer/Deobjectivizer.h>\n";


}

// Top Level driver code
void Deobjectivizer::HandleTopLevelSingleDecl(clang::Decl* D) {
  if(diag.hasErrorOccurred()) return;
  // Two cases: either the decl could be in the main file, or it could be in a
  // #included file.  If the former, rewrite it now.  If the later, check to see
  // if we rewrote the #include/#import.
  clang::SourceLocation Loc = D->getLocation();
  Loc = sm->getInstantiationLoc(Loc);

  // If this is for a builtin, ignore it.
  if(Loc.isInvalid()) return;

  // If this isn't in the main file, ignore it.
  if(!sm->isFromMainFile(D->getLocation())) return;

  // Look for built-in declarations that we need to refer during the rewrite.
  if(clang::FunctionDecl *FD = dyn_cast<clang::FunctionDecl>(D))
    RewriteFunctionDecl(FD);
  else if(clang::ObjCInterfaceDecl* MD =
            dyn_cast<clang::ObjCInterfaceDecl>(D))
    RewriteInterfaceDecl(MD);
  else if(clang::ObjCCategoryDecl* CD =
            dyn_cast<clang::ObjCCategoryDecl>(D))
    RewriteCategoryDecl(CD);
  else if(clang::ObjCProtocolDecl* PD =
            dyn_cast<clang::ObjCProtocolDecl>(D))
    RewriteProtocolDecl(PD);
  else if(clang::ObjCForwardProtocolDecl* FPD =
            dyn_cast<clang::ObjCForwardProtocolDecl>(D))
    RewriteForwardProtocolDecl(FPD);
  else if(clang::ObjCClassDecl* CD =
            dyn_cast<clang::ObjCClassDecl>(D))
    RewriteForwardClassDecl(CD);
  // Look at that variable name :)
  // (Not that I do drugs, but...)
  else if(clang::LinkageSpecDecl* LSD =
            dyn_cast<clang::LinkageSpecDecl>(D)) {
    // Recurse into linkage specifications
    for(clang::DeclContext::decl_iterator DI = LSD->decls_begin(),
        DIEnd = LSD->decls_end();
        DI != DIEnd;
        ++DI)
      HandleTopLevelSingleDecl(*DI);
  }
}

void Deobjectivizer::HandleTranslationUnit(clang::ASTContext& C) {
  if(diag.hasErrorOccurred()) return;

  RewriteInclude();

  InsertText(sm->getLocForStartOfFile(main_id), preamble.c_str(),
             preamble.length(), false);

  // Get the buffer for the main file.
  if(const clang::RewriteBuffer* rewrite_buf =
       rewriter.getRewriteBufferFor(main_id))
    *out_file << std::string(rewrite_buf->begin(), rewrite_buf->end());
  out_file->flush();
  // Now for the implementation file.
  // If it's empty, don't write anything.
  if(!out_c_buf->str().empty()) {
    *out_c_file << "#include \"" << llvm::sys::path::filename(in_file_name) << "\"\n"
                << out_c_buf->str();
    out_c_file->flush();
  }
}

// Syntactic (non-AST) Rewriting code
void Deobjectivizer::RewriteInclude(void) {
  clang::SourceLocation start = sm->getLocForStartOfFile(main_id);
  static const char import[] = "import";
  size_t import_len = sizeof(import)-1;
  static const char include[] = "include";
  size_t include_len = sizeof(include)-1;

  // Loop over the whole file, looking for includes.
  for(const char* buf = main_start; buf < main_end; ++buf) {
    if(*buf == '#') {
      if(++buf == main_end)
        return;
      while(*buf == ' ' || *buf == '\t')
        if(++buf == main_end)
          return;
      if(std::strncmp(buf, import, import_len) == 0) {
        // Replace import with include
        clang::SourceLocation import_loc =
        start.getFileLocWithOffset(buf-main_start);
        ReplaceText(import_loc, import_len, include, include_len);
        buf += import_len;
        // If it's a framework-style include, and the framework
        // is THIS framework, we have to replace the framework
        // name with the equivalent C framework.
        // (e.g. Foundation -> FoundationC)
        // We don't do it for other frameworks because they may or
        // may not be Objective-C frameworks
        // (e.g. CoreFoundation -> CoreFoundationC (bad)).
        RewriteFrameworkInclude(buf, start);
      }
      if(std::strncmp(buf, include, include_len) == 0) {
        buf += include_len;
        // Same here.
        RewriteFrameworkInclude(buf, start);
      }
    }
  }
}
// This is a list of frameworks known to contain Objective-C code.
// If you find a framework whose ONLY interface is Objective-C, add it
// here.
// Note that we explicity exclude frameworks with C interfaces. We're
// not interested in frameworks that don't need to be Deobjectivized.
static const char* known_frameworks[] = {
  "AppKit",
  "Automator",
  "CalendarStore",
  "Cocoa",
  "Collaboration",
  "CoreAudioKit",
  "CoreData",
  "CoreLocation",
  "CoreVideo",
  "CoreWLAN",
  "DiscRecording",
  "DiscRecordingUI",
  "ExceptionHandling",
  "Foundation",
  "ImageCaptureCore",
  "InputMethodKit",
  "InstallerPlugins",
  "InstantMessage",
  "OSAKit",
  "PreferencePanes",
  "PubSub",
  "QTKit",
  "Quartz",
    // Quartz sub-frameworks
    "ImageKit",
    "PDFKit",
    "QuartzComposer",
    "QuartzFilters",
    "QuickLookUI",
  "QuartzCore",
  "QuickLook",
  "RubyCocoa",
  "Scripting",
  "ScriptingBridge",
  "SecurityInterface",
  "ServerNotification",
  "SyncServices",
  "WebKit",   // C interface is limited :(
  "XgridFoundation"
};
void Deobjectivizer::RewriteFrameworkInclude(const char*& buf, clang::SourceLocation start) {
  while(*buf == ' ' || *buf == '\t')
    if(++buf == main_end)
      return;
  // Framework-style includes always start with a '<'.
  if(*buf++ == '<') {
    if(std::strncmp(buf, fwk_name.c_str(),
                    fwk_name.length()) == 0) {
      ReplaceText(start.getFileLocWithOffset(buf-main_start-1), 1, "\"", 1);
      clang::SourceLocation name_loc =
      start.getFileLocWithOffset(buf-main_start);
      RemoveText(name_loc, fwk_name.length()+1);
      const char* rightBracket;
      if((rightBracket = std::strchr(buf, '>')))
        ReplaceText(start.getFileLocWithOffset(rightBracket-main_start), 1, "\"", 1);
    } else {
      for(int i = 0;
          i < sizeof(known_frameworks)/sizeof(known_frameworks[0]);
          ++i) {
        if(std::strncmp(buf, known_frameworks[i],
                        std::strlen(known_frameworks[i])) == 0) {
          ReplaceText(start.getFileLocWithOffset(buf-main_start-1), 1, "\"", 1);
          clang::SourceLocation name_loc =
            start.getFileLocWithOffset(buf-main_start+std::strlen(known_frameworks[i]));
          RemoveText(name_loc, std::strlen(known_frameworks[i])+1);
          const char* rightBracket;
          if((rightBracket = std::strchr(buf, '>')))
            ReplaceText(start.getFileLocWithOffset(rightBracket-main_start), 1, "\"", 1);
          return;
        }
      }
    }
  }
}

void Deobjectivizer::RewriteForwardClassDecl(clang::ObjCClassDecl* decl) {
  // Get the start location and compute the "semi-location".
  clang::SourceLocation start = decl->getLocation();
  const char* start_buf = sm->getCharacterData(start);
  const char* semi_ptr = std::strchr(start_buf, ';');

  // Translate to typedef's that forward reference structs with the
  // same name as the class. As a convenience, we include the original
  // declaration as a comment.
  std::string typedef_str = "/* @class ";
  for(clang::ObjCClassDecl::iterator I = decl->begin(), E = decl->end();
      I != E;
      ++I) {
    clang::ObjCInterfaceDecl* fwd_decl = I->getInterface();
    typedef_str += fwd_decl->getNameAsString();
    if(I+1 != E)
      typedef_str += ", ";
    else
      typedef_str += "; */\n";
  }

  // Now we have to iterate again for the actual typedef's.
  for(clang::ObjCClassDecl::iterator I = decl->begin(), E = decl->end();
      I != E;
      ++I) {
    clang::ObjCInterfaceDecl* fwd_decl = I->getInterface();
    // Protect against multiple defintions.
    typedef_str += "#ifndef _DEOBJ_typedef_";
    typedef_str += fwd_decl->getNameAsString();
    typedef_str += "\n#define _DEOBJ_typedef_";
    typedef_str += fwd_decl->getNameAsString();
    typedef_str += "\ntypedef struct objc_object ";
    typedef_str += fwd_decl->getNameAsString();
    typedef_str += ";\n#endif\n";
  }

  // Now replace the @class declaration with typedefs corresponding to the
  // classes.
  ReplaceText(start, semi_ptr-start_buf+1, typedef_str.c_str(),
              typedef_str.length());
}

void Deobjectivizer::RewriteMethodDeclaration(clang::ObjCMethodDecl* meth) {
  // When the method is a synthesized one... we still have to rewrite it.
  // C clients need to access them too.

  clang::SourceLocation end = meth->getLocEnd();
  if(!meth->isSynthesized())
    end = end.getFileLocWithOffset(1);

  // Rewriting this into a function is somewhat tricky. What we want to do is
  // turn this:
  //  @interface SomeObject {}
  //  ...
  //  - doThis:(int)what withThat:object;
  //  + doSomethingElse;
  //  ...
  //  @end
  // into this:
  //  struct SomeObject {};
  //  ...
  //  id SomeObjectDoThisWithThat(SomeObject* self, int what, id object);
  //
  // In addition, we have to deal with instance vs. class methods.
  // With class methods, we don't want the self parameter:
  //  id SomeObjectDoSomethingElse(void);
  //  ...
  std::string func_decl;
  std::string func_type;
  // And now we need to construct a function declaration.
  // First the return type.
  const clang::FunctionType* fp_ret_type = NULL;
  // We can use the getAsString() method on the type class to write the
  // the result. But there are some caveats.
  // - If the type is a function-pointer type that is NOT a typedef,
  //   then we have to write the type all around, like so:
  //    - (void (*)())someMethod;
  //                |
  //                V
  //    void (*SomeClassSomeMethod(SomeClass* self))();
  // - If the type is a qualified object pointer (e.g., id<NSCoding>), then
  //   we have to write the part NOT inside the angle brackets.
  clang::QualType ret_type = meth->getResultType();
  if(ret_type->isObjCQualifiedIdType())
    func_decl = func_type = "id";
  else if(ret_type->isFunctionPointerType() ||
          ret_type->isBlockPointerType()) {
    clang::QualType pointee_type;
    bool is_block = false;
    if(const clang::PointerType* pt = ret_type->getAs<clang::PointerType>())
       pointee_type = pt->getPointeeType();
    else if(const clang::BlockPointerType* bpt =
            ret_type->getAs<clang::BlockPointerType>()) {
       pointee_type = bpt->getPointeeType();
       is_block = true;
    }
    if((fp_ret_type = pointee_type->getAs<clang::FunctionType>())) {
      // Will break if this is a pointer to a pointer to a function,
      // or a pointer to a pointer to a pointer to a function, etc.
      // But those are so rare, it's simpler right now to just do
      // this.
      func_decl = func_type = fp_ret_type->getResultType().getAsString();
      if(is_block) {
        func_decl += "(^";
        func_type += "(^";
      } else {
        func_decl += "(*";
        func_type += "(*";
      }
    }
  } else
    func_decl = func_type = meth->getResultType().getAsString();
  func_decl += " ";
  func_type += " (*)";
  // Let's grab the name of the class and the name of the
  // method (everything that doesn't name a parameter, including the colons
  // but excluding the sign).
  // There are actually two cases we have to handle. The first case is a
  // normal method, and the second is a protocol method.
  clang::ObjCProtocolDecl* pd =
  dyn_cast<clang::ObjCProtocolDecl>(meth->getDeclContext());
  if(pd)
    func_decl += pd->getName();
  else
    func_decl += meth->getClassInterface()->getName();
  func_decl += "_";
  // If this is a class method, put 'class' in the name.
  if(!meth->isInstanceMethod())
    func_decl += "Class_";
  std::string meth_name = meth->getSelector().getAsString();
  // We need to tokenize the selector name:
  //  doThis:withThat: -> doThis_withThat
  // But since std::string doesn't have any direct tokenization methods,
  // we'll have to do it the hard way: with find() and substr().
  std::string::size_type pos = meth_name.find(':'), old_pos = 0;
  if(pos == std::string::npos) {
    // No colon? Just use the whole string.
    func_decl += meth_name;
  }
  for( ; pos != std::string::npos;
      old_pos = pos+1, pos = meth_name.find(':', old_pos)) {
    // Now append the rest of the fragment.
    func_decl += meth_name.substr(old_pos, pos-old_pos) + "_";
  }
  // Now rewrite the arguments.
  func_decl += "(";
  if(meth->isInstanceMethod())
    func_type += "(id, SEL";
  else
    func_type += "(Class, SEL";

  // First the invisible arguments. We only rewrite 'self'. '_cmd' isn't
  // relevant.
  if(meth->isInstanceMethod()) {
    if(pd) {
      func_decl += "id";
    } else {
      func_decl += meth->getClassInterface()->getName();
      func_decl += " *";
    }
    func_decl += " self";
  } else {
     func_decl += "Class klass";
  }

  // Now for the declared arguments.
  clang::ObjCMethodDecl::param_iterator pi = meth->param_begin();
  if(pi) {
    for(clang::ObjCMethodDecl::param_iterator e = meth->param_end(); pi != e;
        ++pi) {
      clang::ParmVarDecl* pdecl = *pi;
      func_decl += ", ";
      func_type += ", ";
      if(pdecl->getType()->isObjCQualifiedIdType()) {
        // Handle id<SomeProtocol> and friends.
        func_decl += "id ";
        func_decl += pdecl->getName();
        func_type += "id";
      } else {
        std::string name = pdecl->getName();
        pdecl->getType().getAsStringInternal(name, ctx->PrintingPolicy);
        func_decl += name;
        func_type += pdecl->getType().getAsString();
      }
    }
  }
  // If this is a variadic method, make the function variadic.
  if(meth->isVariadic()) {
    func_decl += ", ...";
    func_type += ", ...";
  }
  func_decl += ")";
  func_type += ")";

  if(fp_ret_type) {
    func_decl += ")"; // Close the precedence 'scope' for the return type.
    func_type += ")"; // Same here.
    // Now emit the argument types for the return type (if any).
    if(const clang::FunctionProtoType* FPT =
         dyn_cast<clang::FunctionProtoType>(fp_ret_type)) {
      func_decl += "(";
      func_type += "(";
      for(unsigned i = 0, e = FPT->getNumArgs(); i != e; ++i) {
        if(i) { func_decl += ", "; func_type += ", "; }
        func_decl += FPT->getArgType(i).getAsString();
        func_type += FPT->getArgType(i).getAsString();
      }
      if(FPT->isVariadic()) {
        if(FPT->getNumArgs()) { func_decl += ", "; func_type += ", "; }
        func_decl += "...";
        func_type += "...";
      }
      func_decl += ")";
      func_type += ")";
    } else {
      func_decl += "()";
      func_type += "()";
    }
  }

  // Write the header to the implementation file, first.
  *out_c_buf << func_decl;

  // There. NOW we can terminate the declaration.
  func_decl += ";";
  func_decl = "\n" + func_decl;

  // For convenience, let's add a comment showing the original
  // method declaration.
  // Don't do this if the method was synthesized; there is nothing to
  // comment in that case.
  if(!meth->isSynthesized()) {
    clang::SourceLocation start = meth->getLocStart();

    if(sm->getInstantiationLineNumber(end) >
       sm->getInstantiationLineNumber(start)) {
      InsertText(start, "#if 0\n", 6);
      func_decl = "\n#endif" + func_decl;
    } else {
      InsertText(start, "/* ", 3);
      func_decl = " */" + func_decl;
    }
  }

  // Now insert the function declaration.
  InsertText(end, func_decl.c_str(),
             func_decl.length());

  // Not done yet. Now we have to write the implementation.
  // For our instance method example, the implementation should look
  // like this:
  //  id SomeObjectDoThisWithThat(SomeObject* self, int what, id object) {
  //    return [self doThis:what withThat:object];
  //  }
  // For class methods, we have to do:
  //  id SomeObjectDoSomethingElse(void) {
  //    return [SomeObject doSomethingElse];
  //  }
  *out_c_buf << " {\n  ";
  // If the method doesn't return void, emit a return.
  if(!ret_type->isVoidType())
    *out_c_buf << "return ";
  *out_c_buf << "((" << func_type << ")";
  // If the method returns a struct, use objc_msgSend_stret, else
  // use objc_msgSend. Depending on the target, use objc_msgSend_fpret
  // for floats.
  if(ret_type->isRecordType())
    *out_c_buf << "objc_msgSend_stret";
  else if(ret_type->isFloatingType()) {
    clang::TargetInfo::RealType rt;
    if(ret_type == ctx->FloatTy)
      rt = clang::TargetInfo::Float;
    else if(ret_type == ctx->LongDoubleTy)
      rt = clang::TargetInfo::LongDouble;
    else
      rt = clang::TargetInfo::Double;
    if(ctx->Target.useObjCFPRetForRealType(rt))
      *out_c_buf << "objc_msgSend_fpret";
    else
      *out_c_buf << "objc_msgSend";
  } else
    *out_c_buf << "objc_msgSend";
  *out_c_buf << ")(";
  // If this is a class method, we have to grab the class.
  if(!meth->isInstanceMethod())
    *out_c_buf << "klass == Nil ? (Class)objc_getClass(\"" << meth->getClassInterface()->getName() << "\") : klass";
  else
    *out_c_buf << "self";
  *out_c_buf << ", sel_registerName(\"" << meth_name << "\")";

  // Emit arguments.
  if(meth->param_size() != 0) {
    pi = meth->param_begin();
    pos = 0;
    for(clang::ObjCMethodDecl::param_iterator e = meth->param_end();
        pi != e; ++pi) {
      *out_c_buf << ", " << (*pi)->getName();
    }
  }
  // TODO: Varargs methods.
  *out_c_buf << ");\n}\n\n";
}

void Deobjectivizer::RewriteProperty(clang::ObjCPropertyDecl* prop) {
  // We'll rewrite the getter and setter in RewriteMethodDeclaration()
  // above, so let's just comment out the property.
  clang::SourceLocation start = prop->getAtLoc();

  // Go search for the terminating ';'.
  const char* start_buf = sm->getCharacterData(start);
  const char* end_buf = std::strchr(start_buf, ';');

  // Now we can comment out the property.
  clang::SourceLocation end = start.getFileLocWithOffset(end_buf-start_buf+1);

  if(sm->getInstantiationLoc(start) <
     sm->getInstantiationLoc(end)) {
    InsertText(start, "#if 0\n", 6);
    InsertText(end, "\n#endif", 7);
  } else {
    InsertText(start, "/* ", 3);
    InsertText(end, " */", 3);
  }
}

void Deobjectivizer::RewriteAtEnd(clang::ObjCContainerDecl* cont) {
  clang::SourceRange atend_range = cont->getAtEndRange();
  clang::SourceLocation atend_begin = atend_range.getBegin(),
  atend_end = atend_range.getEnd();
  InsertText(atend_begin, "/* ", 3);
  InsertText(atend_end.getFileLocWithOffset(3), " */", 3);
}

void Deobjectivizer::RewriteCategoryDecl(clang::ObjCCategoryDecl* cat) {
  // First the header.
  clang::SourceLocation start = cat->getAtLoc(),
    catname = cat->getCategoryNameLoc();
  // We have to find the end parenthesis.
  const char* start_buf = sm->getCharacterData(catname);
  const char* end_name_buf = std::strchr(start_buf, ')');
  // If there's protocols involved, we have to find the end bracket
  // of the protocol list now.
  if(cat->protocol_size() > 0)
    end_name_buf = std::strchr(end_name_buf, '>');
  clang::SourceLocation hdr_end =
    catname.getFileLocWithOffset(end_name_buf-start_buf+1);
  if(sm->getInstantiationLoc(start) < sm->getInstantiationLoc(hdr_end)) {
    InsertText(start, "#if 0\n", 6);
    InsertText(hdr_end, "\n#endif", 7);
  } else {
    InsertText(start, "/* ", 3);
    InsertText(hdr_end, " */", 3);
  }

  // Announce in the implementation file that the methods
  // belong to a new category.
  *out_c_buf << "/* " << cat->getClassInterface()->getName() << " methods ("
             << cat->getName() << ") */\n\n";

  // Now rewrite all the properties and method declarations.
  for(clang::ObjCCategoryDecl::prop_iterator i = cat->prop_begin(),
      e = cat->prop_end(); i != e; ++i)
    RewriteProperty(*i);
  for(clang::ObjCCategoryDecl::instmeth_iterator
      i = cat->instmeth_begin(), e = cat->instmeth_end();
      i != e; ++i)
    RewriteMethodDeclaration(*i);
  for(clang::ObjCCategoryDecl::classmeth_iterator
      i = cat->classmeth_begin(), e = cat->classmeth_end();
      i != e; ++i)
    RewriteMethodDeclaration(*i);
    
  // Finally, comment out the @end.
  RewriteAtEnd(cat);
}

void Deobjectivizer::RewriteProtocolDecl(clang::ObjCProtocolDecl* proto) {
  // First the header.
  clang::SourceLocation start = proto->getLocStart();
  clang::SourceLocation end = proto->getLocEnd();
  if(end.isValid())
    end = end.getFileLocWithOffset(1);
  else {
    const char* start_buf = sm->getCharacterData(start);
    const char* end_buf = std::strchr(start_buf, '\n');
    if(end_buf) end = start.getFileLocWithOffset(end_buf-start_buf);
  }
  if(sm->getInstantiationLoc(start) < sm->getInstantiationLoc(end)) {
    InsertText(start, "#if 0\n", 6);
    InsertText(end, "\n#endif", 7);
  } else {
    InsertText(start, "/* ", 3);
    InsertText(end, " */", 3);
  }

  // Announce in the implementation file that the methods
  // belong to a new protocol.
  *out_c_buf << "/* " << proto->getName() << " methods */\n\n";

  // Now rewrite all the method declarations.
  for(clang::ObjCProtocolDecl::instmeth_iterator
      i = proto->instmeth_begin(), e = proto->instmeth_end();
      i != e; ++i)
    RewriteMethodDeclaration(*i);
  for(clang::ObjCProtocolDecl::classmeth_iterator
      i = proto->classmeth_begin(), e = proto->classmeth_end();
      i != e; ++i)
    RewriteMethodDeclaration(*i);

  // Finally, comment out the @end.
  RewriteAtEnd(proto);

  // We're not done yet! Now we have to comment out @optional and @required.
  // Note that there can be whitespace and comments between the '@' and the
  // optional/required.
  const char* start_buf = sm->getCharacterData(start);
  const char* end_buf = sm->getCharacterData(proto->getAtEndRange().getBegin());
  for(const char* p = start_buf; p < end_buf; ++p) {
    if(*p == '@') {
      // First the common cases.
      static const char optional[] = "optional";
      static const char required[] = "required";
      if(!std::strncmp(p+1, optional, sizeof(optional)-1)) {
        const char commented_optional[] = "/* @optional */";
        clang::SourceLocation optional_loc =
        start.getFileLocWithOffset(p-start_buf);
        ReplaceText(optional_loc, sizeof(optional),
                    commented_optional, sizeof(commented_optional)-1);
      } else if(!std::strncmp(p+1, required, sizeof(required)-1)) {
        const char commented_required[] = "/* @required */";
        clang::SourceLocation required_loc =
        start.getFileLocWithOffset(p-start_buf);
        ReplaceText(required_loc, sizeof(required),
                    commented_required, sizeof(commented_required)-1);
      }
      // FIXME: Handle the other cases!
      // Not really important right now, because most ObjC keywords
      // are written with no whitespace between the @ and the name.
      // So constructs like these:
      //  @
      //    // Here's a comment! Forget you, deobjectivizer!
      //   optional
      // are very rare.
    }
  }
}

void Deobjectivizer::RewriteForwardProtocolDecl(clang::ObjCForwardProtocolDecl* proto) {
  clang::SourceLocation start = proto->getLocStart();
  const char* start_buf = sm->getCharacterData(start);
  const char* end_buf = std::strchr(start_buf, ';');
  clang::SourceLocation end = start.getFileLocWithOffset(end_buf-start_buf+1);
  assert(!start.isInvalid() && "Invalid SourceLocation");
  // We "handle foward protocol declared across multiple lines."
  if(sm->getInstantiationLoc(start) <
     sm->getInstantiationLoc(end)) {
    InsertText(start, "#if 0\n", 6);
    InsertText(end, "\n#endif", 7);
  } else {
    InsertText(start, "/* ", 3);
    InsertText(end, " */", 3);
  }
}

void Deobjectivizer::RewriteInterfaceDecl(clang::ObjCInterfaceDecl* iface) {
  std::string result_str;
  // Generate a forward decl typedef.
  result_str = "#ifndef _DEOBJ_typedef_";
  result_str += iface->getName();
  result_str += "\n#define _DEOBJ_typedef_";
  result_str += iface->getName();
  result_str += "\ntypedef struct objc_object ";
  result_str += iface->getName();
  result_str += ";\n#endif\n\n";
  // Clients that need to access the public members can use the i-var
  // access functions, so just comment out the struct.
  clang::SourceLocation start = iface->getLocStart();
  clang::SourceLocation end = iface->getLocEnd().getFileLocWithOffset(1);
  const char* start_buf = sm->getCharacterData(start);
  const char* end_buf = sm->getCharacterData(end),* cursor = end_buf;
  while(std::isalnum(*cursor) || *cursor == '_')
    cursor++;
  end = end.getFileLocWithOffset(cursor-end_buf);
  if(BufferContainsEndIfPPDirective(start_buf, end_buf)) {
    cursor = start_buf;
    // Back up to the #if (or #ifdef or #ifndef)
    while(!BufferContainsIfPPDirective(start_buf, end_buf) &&
          !BufferContainsIfDefPPDirective(start_buf, end_buf) &&
          !BufferContainsIfNdefPPDirective(start_buf, end_buf))
      --start_buf;
    start = start.getFileLocWithOffset(start_buf-cursor);
  }
  if(sm->getInstantiationLoc(start) < sm->getInstantiationLoc(end)) {
    result_str += "#if 0\n";
    InsertText(end, "\n#endif", 7);
  } else {
    result_str += "/* ";
    InsertText(end, " */", 3);
  }
  InsertText(start, result_str.c_str(), result_str.length());

  // Announce in the implementation file that the methods
  // belong to a new class.
  *out_c_buf << "/* " << iface->getName() << " methods */\n\n";

  // Now rewrite the properties and methods.
  for(clang::ObjCInterfaceDecl::prop_iterator i = iface->prop_begin(),
      e = iface->prop_end(); i != e; ++i)
    RewriteProperty(*i);
  for(clang::ObjCInterfaceDecl::instmeth_iterator i = iface->instmeth_begin(),
      e = iface->instmeth_end(); i != e; ++i)
    RewriteMethodDeclaration(*i);
  for(clang::ObjCInterfaceDecl::classmeth_iterator
      i = iface->classmeth_begin(), e = iface->classmeth_end(); i != e; ++i)
    RewriteMethodDeclaration(*i);

  // Lastly, comment out the @end.
  RewriteAtEnd(iface);
}

static bool needToScanForQualifiers(clang::QualType type) {
  if(type->isObjCQualifiedIdType())
    return true;
  if(const clang::PointerType* pt = type->getAs<clang::PointerType>())
    if(pt->getPointeeType()->isObjCQualifiedIdType())
      return true;
  if(type->isObjCObjectPointerType())
    return type->getPointeeType()->isObjCQualifiedInterfaceType();
  return false;
}
static bool scanForProtocolRefs(const char* start, const char* end,
                                const char*& start_ref, const char*& end_ref) {
  for( ; start < end; ++start) {
    if(*start == '<')
      start_ref = start; // Mark the start of the protocol ref.
    if(*start == '>') {
      // Can't have an end without a start.
      if(start_ref && *start_ref == '<') {
        end_ref = start;
        return true;
      }
      return false;
    }
  }
  return false;
}
static void scanToNextArgument(const char*& arg_ref) {
  int angle = 0;
  for( ; *arg_ref != ')' && (*arg_ref != ',' || angle > 0); ++arg_ref) {
    if(*arg_ref == '<')
      angle++;    // Enter some angled brackets.
    else if(*arg_ref == '>')
      angle--;    // Leave the angled brackets.
  }
  assert(angle == 0 && "Protocol reference is unterminated");
}

void Deobjectivizer::RewriteObjCQualifiedInterfaceTypes(clang::Expr* expr) {
  clang::QualType type = expr->getType();
  if(needToScanForQualifiers(type)) {
    clang::SourceLocation start, end;

    if(const clang::CStyleCastExpr* ce =
       dyn_cast<clang::CStyleCastExpr>(expr)) {
      start = ce->getLParenLoc();
      end = ce->getRParenLoc();
    } else {
      start = expr->getLocStart();
      end = expr->getLocEnd();
    }
    // Don't even try to rewrite stuff we synthesized ourselves.
    // Shouldn't get here, but you never know...
    if(start.isInvalid() || end.isInvalid())
      return;

    const char* start_buf = sm->getCharacterData(start);
    const char* end_buf = sm->getCharacterData(end);
    const char* start_ref = NULL,* end_ref = NULL;
    if(scanForProtocolRefs(start_buf, end_buf, start_ref, end_ref)) {
      // Get the SourceLocations of the angle brackets.
      clang::SourceLocation open_loc =
        start.getFileLocWithOffset(start_ref-start_buf);
      clang::SourceLocation close_loc =
        start.getFileLocWithOffset(end_ref-start_buf+1);
      // Now comment out the protocol references.
      InsertText(open_loc, "/*", 2);
      InsertText(close_loc, "*/", 2);
    }
  }
}
void Deobjectivizer::RewriteObjCQualifiedInterfaceTypes(clang::Decl* decl) {
  clang::SourceLocation loc;
  clang::QualType type;
  const clang::FunctionProtoType* proto_type = NULL;
  if(clang::VarDecl* vd = dyn_cast<clang::VarDecl>(decl)) {
    // Ordinary variable.
    loc = vd->getLocation();
    type = vd->getType();
  } else if(clang::FunctionDecl* fd = dyn_cast<clang::FunctionDecl>(decl)) {
    // Function that might return a qualified pointer.
    loc = fd->getLocation();
    // Make sure we're really dealing with a function.
    const clang::FunctionType* ft =
      fd->getType()->getAs<clang::FunctionType>();
    assert(ft && "Function decl without function type");
    // While we're at it, we may as well check the function parameters.
    proto_type = dyn_cast<clang::FunctionProtoType>(ft);
    type = ft->getResultType();
  } else if(clang::FieldDecl* fd = dyn_cast<clang::FieldDecl>(decl)) {
    // Field in a structure, class, union; or, an i-var.
    loc = fd->getLocation();
    type = fd->getType();
  } else
    // We're only interested in decls that have types associated with them.
    return;

  if(needToScanForQualifiers(type)) {
    // Time to scan the buffer.
    const char* end = sm->getCharacterData(loc);
    const char* start = end;
    while(*start != ';' && *start != '<' && start != main_start)
      --start;    // scan backward from the decl loc for return types.
    const char* start_ref = NULL,* end_ref = NULL;
    if(scanForProtocolRefs(start, end, start_ref, end_ref)) {
      // Get the SourceLocations of the brackets.
      clang::SourceLocation open_loc = loc.getFileLocWithOffset(start_ref-end);
      clang::SourceLocation close_loc = loc.getFileLocWithOffset(end_ref-end+1);
      // Now comment out the protocol references.
      InsertText(open_loc, "/*", 2);
      InsertText(close_loc, "*/", 2);
    }
  }
  // If this isn't a function type with a prototype, we're done!
  if(!proto_type) return;
  // Now do the arguments.
  const char* start = sm->getCharacterData(loc);
  const char* start_func = start;
  for(unsigned i = 0; i < proto_type->getNumArgs(); ++i) {
    if(needToScanForQualifiers(proto_type->getArgType(i))) {
      // Once again, we must scan the buffer.
      const char* end = start;
      // Scan forward from the decl loc for argument types.
      scanToNextArgument(end);
      const char* start_ref = NULL,* end_ref = NULL;
      if(scanForProtocolRefs(start, end, start_ref, end_ref)) {
        // Get the SourceLocations of the brackets.
        clang::SourceLocation open_loc =
          loc.getFileLocWithOffset(start_ref-start_func);
        clang::SourceLocation close_loc =
          loc.getFileLocWithOffset(end_ref-start_func+1);
        // Comment out the protocol references.
        InsertText(open_loc, "/*", 2);
        InsertText(close_loc, "*/", 2);
      }
      start = ++end;
    } else {
      // From RewriteObjC: "If the function name is derived from a macro
      // expansion, then the argument buffer will not follow the name."
      // We'll need help with this...
    }
  }
}

void Deobjectivizer::RewriteFunctionDecl(clang::FunctionDecl* func) {
  // From <objc/objc.h> (and <Deobjectivizer/Deobjectivizer.h> :).
  if(func->getIdentifier() && func->getName() == "sel_registerName") {
    SelRegisterNameFunctionDecl = func;
    return;
  }
  // The only thing we really need to rewrite are the arguments and
  // return type, in case they have qualified object pointers.
  RewriteObjCQualifiedInterfaceTypes(func);
}

// Expression rewriting.
clang::Stmt* Deobjectivizer::RewriteAtEncode(clang::ObjCEncodeExpr* expr) {
  // Create a new string literal expression.
  clang::QualType str_type = ctx->getPointerType(ctx->CharTy);
  std::string encoding;
  ctx->getObjCEncodingForType(expr->getEncodedType(), encoding);
  clang::Expr* replacement = clang::StringLiteral::Create(*ctx,
                                                          encoding.c_str(),
                                                          encoding.length(),
                                                          false, str_type,
                                                          clang::SourceLocation());
  ReplaceStmt(expr, replacement);

  //delete expr; // Leak for now...
  return replacement;
}

clang::Stmt* Deobjectivizer::RewriteAtSelector(clang::ObjCSelectorExpr* expr) {
  // Use the sel_getUid() function to turn the method name into an
  // actual selector.
  // (Actually it's the sel_registerName() function, since sel_getUid()
  // is deprecated.)
  if(!SelRegisterNameFunctionDecl)
    SynthSelRegisterNameFunctionDecl();
  assert(SelRegisterNameFunctionDecl && "Can't find sel_registerName() decl");
  // Now call sel_registerName().
  llvm::SmallVector<clang::Expr*, 4> sel_exprs;
  clang::QualType arg_type = ctx->getPointerType(ctx->CharTy);
  sel_exprs.push_back(clang::StringLiteral::Create(*ctx,
                                                   expr->getSelector().getAsString().c_str(),
                                                   expr->getSelector().getAsString().length(),
                                                   false, arg_type, clang::SourceLocation()));
  clang::CallExpr* sel_exp = SynthesizeCallToFunctionDecl(SelRegisterNameFunctionDecl,
                                                          &sel_exprs[0],
                                                          sel_exprs.size());
  ReplaceStmt(expr, sel_exp);
  //delete expr; // Leak for now...
  return sel_exp;
}

// Syntax and expression synthesis.
clang::CallExpr* Deobjectivizer::SynthesizeCallToFunctionDecl(clang::FunctionDecl* func,
                                                              clang::Expr** args,
                                                              unsigned nargs) {
  // Get the function type, we need it a couple of places.
  clang::QualType func_type = func->getType();

  // Create a reference to the FunctionDecl.
  clang::DeclRefExpr* dre = new (ctx) clang::DeclRefExpr(func, func_type,
                                                         clang::VK_LValue,
                                                         clang::SourceLocation());

  // Now we cast the reference to a pointer to the function type.
  clang::QualType func_ptr_type = ctx->getPointerType(func_type);
  clang::ImplicitCastExpr* ice = clang::ImplicitCastExpr::Create(
    *ctx, func_ptr_type, clang::CK_FunctionToPointerDecay, dre, 0,
    clang::VK_LValue);

  const clang::FunctionType* ft = func_type->getAs<clang::FunctionType>();

  return new (ctx) clang::CallExpr(*ctx, ice, args, nargs,
                                   ft->getResultType(), clang::VK_RValue,
                                   clang::SourceLocation());
}

void Deobjectivizer::SynthSelRegisterNameFunctionDecl(void) {
  // SEL sel_registerName(const char* str);
  clang::IdentifierInfo* srn_id = &ctx->Idents.get("sel_registerName");
  llvm::SmallVector<clang::QualType, 16> arg_types;
  arg_types.push_back(ctx->getPointerType(ctx->CharTy.withConst()));
  clang::FunctionProtoType::ExtProtoInfo epi;
  clang::QualType get_func_type = ctx->getFunctionType(ctx->getObjCSelType(),
                                                       &arg_types[0],
                                                       arg_types.size(),
                                                       clang::FunctionProtoType::ExtProtoInfo());
  clang::TypeSourceInfo* tinfo = ctx->getTrivialTypeSourceInfo(get_func_type,
                                                               clang::SourceLocation());
  SelRegisterNameFunctionDecl = clang::FunctionDecl::Create(*ctx, tu_decl,
                                                            clang::SourceLocation(),
                                                            clang::SourceLocation(),
                                                            clang::DeclarationName(srn_id),
                                                            get_func_type,
                                                            tinfo,
                                                            clang::SC_Extern);
}

// Miscellaneous utility.
static bool BufferContainsPPDirective(const char* start, const char* end,
                                      const char* which, size_t len) {
    for( ; start < end; ++start) {
        if(*start == '#') {
            // Skip whitespace.
            for(++start; *start == ' ' || *start == '\t'; ++start)
                ;
            if (!std::strncmp(start, which, len))
                return true;
        }
    }
    return false;
}
bool Deobjectivizer::BufferContainsIfPPDirective(const char* start,
                                                 const char* end) {
    return BufferContainsPPDirective(start, end, "if", sizeof("if")-1);
}
bool Deobjectivizer::BufferContainsIfDefPPDirective(const char* start,
                                                    const char* end) {
    return BufferContainsPPDirective(start, end, "ifdef", sizeof("ifdef")-1);
}
bool Deobjectivizer::BufferContainsIfNdefPPDirective(const char* start,
                                                     const char* end) {
    return BufferContainsPPDirective(start, end, "ifndef", sizeof("ifndef")-1);
}
bool Deobjectivizer::BufferContainsEndIfPPDirective(const char* start,
                                                    const char* end) {
    return BufferContainsPPDirective(start, end, "endif", sizeof("endif")-1);
}

