// Deobjectivizer - The program that makes C bindings to ObjC frameworks
// Copyright (C) 2010 Charles E. Davis

// This program is free software...

// DeobjectivizerInvocation.cpp - Implements the DeobjectivizerInvocation class.

#include <set>
#include <llvm/Support/system_error.h>
#include <llvm/Support/PathV2.h>
#include <llvm/Support/FileSystem.h>
#include <clang/Frontend/FrontendOptions.h>
#include <clang/Frontend/FrontendDiagnostic.h>
#include <clang/Frontend/CompilerInstance.h>
#include "DeobjectivizerInvocation.h"

static void getPathList(llvm::StringRef path, llvm::SmallVectorImpl<std::string> &paths) {
    std::size_t at = 0, delim = path.find(llvm::sys::PathSeparator);
    while (delim != llvm::StringRef::npos) {
        paths.push_back(path.slice(at, delim));
        at = delim+1;
        delim = path.find(llvm::sys::PathSeparator, at);
    }
    if (at != path.size())
        paths.push_back(path.substr(at));
}

void GetSystemFrameworkPaths(llvm::SmallVectorImpl<std::string> &result) {
    if (const char *env = ::getenv("DYLD_FRAMEWORK_PATH")) {
        getPathList(env, result);
    }
    result.push_back("/Library/Frameworks");
    result.push_back("/System/Library/Frameworks");
}

llvm::error_code FindFramework(const llvm::Twine &name, llvm::SmallVectorImpl<char> &result) {
    llvm::SmallVector<std::string, 2> paths;
    GetSystemFrameworkPaths(paths);
    llvm::error_code err;
    // Test each path to see if the framework is in there.
    for (llvm::SmallVectorImpl<std::string>::iterator I = paths.begin(),
         E = paths.end(); I != E; ++I) {
        llvm::Twine path = llvm::Twine(*I) + "/" + name + ".framework";
        bool exist;
        err = llvm::sys::fs::exists(path, exist);
        if (err) return err;
        if (!exist) continue;
        path.toVector(result);
        return llvm::make_error_code(llvm::errc::success);
    }
    return llvm::make_error_code(llvm::errc::no_such_file_or_directory);
}

void DeobjectivizerInvocation::setFrameworkName(const std::string &fwk_name, clang::Diagnostic &diag) {
    DeobjOpts.FrameworkName = fwk_name;
    getFrontendOpts().Inputs.clear();
    llvm::SmallString<256> fwk_path;
    llvm::error_code err;
    if ((err = FindFramework(fwk_name, fwk_path)).value() != llvm::errc::success) {
        diag.Report(clang::diag::err_fe_error_opening) << fwk_name << err.message();
        return;
    }
    // Enumerate all the header files in the framework's Headers directory,
    // except the master header (the one with the same name as the framework
    // itself).
    llvm::sys::path::append(fwk_path, "Headers");
    llvm::sys::fs::directory_iterator it(llvm::Twine(fwk_path), err), e;
    llvm::sys::fs::file_status st;
	if (err.value() != llvm::errc::success) {
		diag.Report(clang::diag::err_fe_error_reading) << fwk_name << err.message();
		return;
	}
    for(; it != e; it.increment(err)) {
        it->status(st);
        if(st.type() == llvm::sys::fs::file_type::status_error ||
           st.type() == llvm::sys::fs::file_type::directory_file ||
           st.type() == llvm::sys::fs::file_type::type_unknown)
            continue;
        if(llvm::sys::path::stem(it->path()) == fwk_name &&
		   llvm::sys::path::extension(it->path()) == ".h")
            continue;
        getFrontendOpts().Inputs.push_back(std::pair<clang::InputKind, std::string>(clang::IK_ObjC, it->path()));
    }
}

