// Deobjectivizer - The program that makes C bindings to ObjC frameworks
// Copyright (C) 2010 Charles E. Davis

// This program is free software...

// main.cpp - Entry point for the Deobjectivizer program.

#include <llvm/ADT/OwningPtr.h>
#include <llvm/ADT/SmallString.h>
#include <llvm/ADT/ArrayRef.h>
#include <llvm/Support/ManagedStatic.h>
#include <llvm/Support/Host.h>
#include <llvm/Support/FileSystem.h>
#include <clang/Driver/Compilation.h>
#include <clang/Driver/Driver.h>
#include <clang/Driver/Tool.h>
#include <clang/Frontend/CompilerInstance.h>
#include <clang/Frontend/FrontendDiagnostic.h>
#include <clang/Frontend/TextDiagnosticPrinter.h>
#include "DeobjectivizerInvocation.h"
#include "DeobjectivizerAction.h"

llvm::sys::Path getMainExe(const char* argv0) {
    return llvm::sys::Path::GetMainExecutable(argv0, (void*)(intptr_t)getMainExe);
}

int main (int argc, const char * const argv[]) {
    void* main_addr = (void*)(intptr_t)getMainExe;
    llvm::sys::Path main_path = getMainExe(argv[0]);
    llvm::IntrusiveRefCntPtr<clang::DiagnosticIDs> diag_id(new clang::DiagnosticIDs);
    DeobjectivizerInvocation *di = new DeobjectivizerInvocation;
    clang::TextDiagnosticPrinter *dpr = new clang::TextDiagnosticPrinter(llvm::errs(), clang::DiagnosticOptions());
    clang::Diagnostic diag(diag_id, dpr);
    clang::driver::Driver driver(main_path.str(), llvm::sys::getHostTriple(),
                                 "a.out", false, false, diag);
    driver.setTitle("deobjectivizer");

    llvm::SmallVector<const char*, 2> args;
    args.push_back(argv[0]);
    args.push_back("-fsyntax-only");
    // FIXME: Hack to get the driver to generate a job for us...
    llvm::SmallString<128> temp_file;
    int fd;
    llvm::sys::fs::unique_file("deobjectivizer-%%%%%%.m", fd, temp_file);
    args.push_back(temp_file.c_str());
    llvm::OwningPtr<clang::driver::Compilation> comp(driver.BuildCompilation(
                                            llvm::ArrayRef<const char*>(args)));
    bool removed;
    llvm::sys::fs::remove(temp_file.str(), removed);
    ::close(fd);
    if(!comp) return 1;

    const clang::driver::JobList& jobs = comp->getJobs();
    if (jobs.size() != 1 || !llvm::isa<clang::driver::Command>(jobs.begin())) {
        llvm::SmallString<256> msg;
        llvm::raw_svector_ostream os(msg);
        comp->PrintJob(os, jobs, "; ", true);
        diag.Report(clang::diag::err_fe_expected_compiler_job) << os.str();
        return 1;
    }

    const clang::driver::Command *cmd = cast<clang::driver::Command>(*jobs.begin());
    if (llvm::StringRef(cmd->getCreator().getName()) != "clang") {
        diag.Report(clang::diag::err_fe_expected_clang_command);
        return 1;
    }

    const clang::driver::ArgStringList &myargs = cmd->getArguments();
    llvm::OwningPtr<clang::CompilerInstance> ci(new clang::CompilerInstance);
    DeobjectivizerAction act;

    clang::CompilerInvocation::CreateFromArgs(*di,
                                       const_cast<const char**>(myargs.data()),
                         const_cast<const char**>(myargs.data())+myargs.size(),
                                              diag);
    ci->setInvocation(di);

    ci->createDiagnostics((int)myargs.size(),const_cast<char**>(myargs.data()));
    if(!ci->hasDiagnostics()) return 1;

    if(ci->getHeaderSearchOpts().UseBuiltinIncludes &&
       ci->getHeaderSearchOpts().ResourceDir.empty())
        ci->getHeaderSearchOpts().ResourceDir =
            clang::CompilerInvocation::GetResourcesPath(argv[0], main_addr);

    llvm::SmallVector<const char*, 4> fwks(argv+1, argv+argc);
    for(llvm::SmallVectorImpl<const char*>::iterator i = fwks.begin(),
        e = fwks.end(); i != e; ++i) {
        di->setFrameworkName(*i, diag);
        if(!diag.hasErrorOccurred())
            ci->ExecuteAction(act);
    }
    llvm::llvm_shutdown();
    return 0;
}
